package org.jboss.cache.interceptors;

import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.Option;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jgroups.Address;

import java.util.List;
import java.util.Map;

/**
 * Takes care of replicating modifications to other nodes in a cluster. Also
 * listens for prepare(), commit() and rollback() messages which are received
 * 'side-ways' (see docs/design/Refactoring.txt).
 *
 * @author Bela Ban
 * @version $Id: ReplicationInterceptor.java 4968 2008-01-03 19:17:09Z manik.surtani@jboss.com $
 */
public class ReplicationInterceptor extends BaseRpcInterceptor
{
   public ReplicationInterceptor()
   {
      initLogger();
   }

   protected boolean skipMethodCall(InvocationContext ctx)
   {
      Option optionOverride = ctx.getOptionOverrides();
      if (optionOverride != null && optionOverride.isCacheModeLocal() && ctx.getTransaction() == null)
      {
         if (trace) log.trace("skip replication");
         return true;
      }
      return false;
   }

   protected Object handleCommitMethod(InvocationContext ctx, GlobalTransaction gtx) throws Throwable
   {
      if (skipReplciationOfTransactionMethod(ctx))
      {
         return nextInterceptor(ctx);
      }
      replicateCall(ctx, ctx.getMethodCall(), configuration.isSyncCommitPhase(), ctx.getOptionOverrides());
      return nextInterceptor(ctx);
   }

   protected Object handlePrepareMethod(InvocationContext ctx, GlobalTransaction gtx, List modification, Address coordinator, boolean onePhaseCommit) throws Throwable
   {
      if (skipReplciationOfTransactionMethod(ctx))
      {
         return nextInterceptor(ctx);
      }
      Object retVal = nextInterceptor(ctx);
      runPreparePhase(ctx.getMethodCall(), gtx, ctx);
      return retVal;
   }

   protected Object handleRollbackMethod(InvocationContext ctx, GlobalTransaction gtx) throws Throwable
   {
      if (skipReplciationOfTransactionMethod(ctx))
      {
         return nextInterceptor(ctx);
      }
      if (!ctx.isLocalRollbackOnly())
      {
         replicateCall(ctx, ctx.getMethodCall(), configuration.isSyncRollbackPhase(), ctx.getOptionOverrides());
      }
      return nextInterceptor(ctx);

   }

   protected Object handlePutForExternalReadMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value) throws Throwable
   {
      if (isTransactionalAndLocal(ctx))
      {
         Object returnValue = nextInterceptor(ctx);
         cache.getTransactionTable().get(gtx).setForceAsyncReplication(true);
         return returnValue;
      }
      else
      {
         return handleCrudMethod(ctx);
      }
   }

   /**
    * It does not make sense replicating a transaction method(commit, rollback, prepare) if one of the following:
    * <pre>
    *    - call was not initiated here, but on other member of the cluster
    *    - there is no transaction. Why calling a commit if no transaction going on?
    *    - the current transaction did not modufy any data, so other members are not aware of it
    * </pre>
    */
   private boolean skipReplciationOfTransactionMethod(InvocationContext ctx)
   {
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      boolean isInitiatedHere = gtx != null && !gtx.isRemote();
      if (trace) log.trace("isInitiatedHere? " + isInitiatedHere + "; gtx = " + gtx);
      return !isTransactionalAndLocal(ctx) || !containsModifications(ctx);
   }

   /**
    * The call runs in a transaction and it was initiated on this node of the cluster.
    */
   private boolean isTransactionalAndLocal(InvocationContext ctx)
   {
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      boolean isInitiatedHere = gtx != null && !gtx.isRemote();
      return isInitiatedHere && (ctx.getTransaction() != null);
   }

   protected Object handlePutDataMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, Map data, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handlePutDataEraseMethod(InvocationContext ctx, GlobalTransaction gt, Fqn fqn, Map newData, boolean createUndoOps, boolean eraseContents) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handlePutKeyValueVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value, boolean createUndoOps, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handlePutDataEraseVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Map data, boolean createUndoOps, boolean eraseContent, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handlePutDataVersionedMethod(InvocationContext ctx, GlobalTransaction globalTransaction, Fqn fqn, Map map, Boolean createUndoOps, DataVersion dataVersion) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handlePutKeyValueMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handlePutForExternalReadVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, Object value, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handleRemoveNodeMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handleRemoveKeyMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, Object key, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handleRemoveDataMethod(InvocationContext ctx, GlobalTransaction tx, Fqn fqn, boolean createUndoOps) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handleDataGravitationCleanupMethod(InvocationContext ctx, Fqn primary, Fqn backup) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handleMoveMethod(InvocationContext ctx, Fqn from, Fqn to) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handleRemoveNodeVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, boolean createUndoOps, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handleRemoveKeyVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, Object key, boolean createUndoOps, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   protected Object handleRemoveDataVersionedMethod(InvocationContext ctx, GlobalTransaction gtx, Fqn fqn, boolean createUndoOps, DataVersion dv) throws Throwable
   {
      return handleCrudMethod(ctx);
   }

   /**
    * If we are within one transaction we won't do any replication as replication would only be performed at commit time.
    * If the operation didn't originate locally we won't do any replication either.
    */
   private Object handleCrudMethod(InvocationContext ctx)
         throws Throwable
   {
      // FIRST pass this call up the chain.  Only if it succeeds (no exceptions) locally do we attempt to replicate.
      Object returnValue = nextInterceptor(ctx);
      if (ctx.getTransaction() == null && ctx.isOriginLocal())
      {
         MethodCall m = ctx.getMethodCall();
         if (trace)
         {
            log.trace("invoking method " + m + ", members=" + cache.getMembers() + ", mode=" +
                  configuration.getCacheMode() + ", exclude_self=" + true + ", timeout=" +
                  configuration.getSyncReplTimeout());
         }
         if (!isSynchronous(ctx.getOptionOverrides()) || m.getMethodId() == MethodDeclarations.putForExternalReadMethodLocal_id)
         {
            // 2. Replicate change to all *other* members (exclude self !)
            replicateCall(ctx, m, false, ctx.getOptionOverrides());
         }
         else
         {
            // REVISIT Needs to exclude itself and apply the local change manually.
            // This is needed such that transient field is modified properly in-VM.
            replicateCall(ctx, m, true, ctx.getOptionOverrides());
         }
      }
      return returnValue;
   }

   /**
    * Calls prepare(GlobalTransaction,List,org.jgroups.Address,boolean)) in all members except self.
    * Waits for all responses. If one of the members failed to prepare, its return value
    * will be an exception. If there is one exception we rethrow it. This will mark the
    * current transaction as rolled back, which will cause the
    * afterCompletion(int) callback to have a status
    * of <tt>MARKED_ROLLBACK</tt>. When we get that call, we simply roll back the
    * transaction.<br/>
    * If everything runs okay, the afterCompletion(int)
    * callback will trigger the @link #runCommitPhase(GlobalTransaction)).
    * <br/>
    *
    * @throws Exception
    */
   protected void runPreparePhase(MethodCall prepareMethod, GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      boolean async = configuration.getCacheMode() == Configuration.CacheMode.REPL_ASYNC;
      if (trace)
      {
         log.trace("(" + cache.getLocalAddress() + "): running remote prepare for global tx " + gtx + " with async mode=" + async);
      }

      // this method will return immediately if we're the only member (because exclude_self=true)
      replicateCall(ctx, prepareMethod, !async, ctx.getOptionOverrides());
   }
}
