package org.jboss.cache.eviction;


import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.lock.IsolationLevel;
import org.jboss.cache.misc.TestingUtil;
import static org.testng.AssertJUnit.*;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * Unit tests for LRU Policy.
 *
 * @author Ben Wang, Feb 11, 2004
 * @author Daniel Huang - dhuang@jboss.org
 * @version $Revision: 4880 $
 */
@Test(groups = {"functional"})
public class LRUPolicyTest
{
   CacheSPI<Object, Object> cache_;
   int wakeupIntervalMillis_ = 0;
   final String ROOT_STR = "/test";
   Throwable t1_ex, t2_ex;
   final long DURATION = 10000;
   boolean isTrue;

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      initCaches();
      wakeupIntervalMillis_ = cache_.getConfiguration().getEvictionConfig().getWakeupIntervalSeconds() * 1000;
      log("wakeupInterval is " + wakeupIntervalMillis_);
      if (wakeupIntervalMillis_ < 0)
      {
         fail("testEviction(): eviction thread wake up interval is illegal " + wakeupIntervalMillis_);
      }

      t1_ex = t2_ex = null;
      isTrue = true;
   }

   public void initCaches() throws Exception
   {
      cache_ = (CacheSPI<Object, Object>) new DefaultCacheFactory().createCache("META-INF/local-lru-eviction-service.xml", false);// read in generic local xml
      cache_.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache_.getConfiguration().setIsolationLevel(IsolationLevel.SERIALIZABLE);
      cache_.start();
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      cache_.stop();
   }

   public void testInUseEviction() throws Exception
   {
      String rootStr = "/org/jboss/test/data/inuse/";
      Fqn fqn;
      //      cache_.put("/", "a", "b");
      //      TestingUtil.sleepThread(wakeupIntervalMillis_ + 500);

      for (int i = 0; i < 10; i++)
      {
         String str = rootStr + i;
         fqn = Fqn.fromString(str);
         cache_.put(fqn, str, str);
      }

      TestingUtil.sleepThread(wakeupIntervalMillis_ + 500);
      System.out.println("***************************** marking as in-use");
      cache_.getRegionManager().getRegion(Fqn.fromString(rootStr + 5), false).markNodeCurrentlyInUse(Fqn.fromString(rootStr + 5), 0);

      for (int i = 10; i < 15; i++)
      {
         String str = rootStr + i;
         fqn = Fqn.fromString(str);
         cache_.put(fqn, str, str);
      }

      TestingUtil.sleepThread(wakeupIntervalMillis_ + 500);

      for (int i = 0; i < 5; i++)
      {
         assertNull(cache_.getNode(Fqn.fromString(rootStr + i)));
      }

      assertNotNull(cache_.getNode(Fqn.fromString(rootStr + 5)));

      for (int i = 6; i < 11; i++)
      {
         assertNull(cache_.getNode(Fqn.fromString(rootStr + i)));
      }
      for (int i = 11; i < 15; i++)
      {
         assertNotNull(cache_.getNode(Fqn.fromString(rootStr + i)));
      }
   }

   public void testEviction()
   {
      String rootStr = "/org/jboss/test/data/";
      for (int i = 0; i < 10; i++)
      {
         String str = rootStr + i;
         Fqn fqn = Fqn.fromString(str);
         try
         {
            cache_.put(fqn, str, str);
         }
         catch (Exception e)
         {
            fail("Failed to insert data" + e);
            e.printStackTrace();
         }
      }
      System.out.println(cache_.toString());
      TestingUtil.sleepThread(wakeupIntervalMillis_ + 500);
      System.out.println(cache_.toString());
      try
      {
         String val = (String) cache_.get(rootStr + "3", rootStr + "3");
         assertNull("DataNode should be empty ", val);
      }
      catch (Exception e)
      {
         e.printStackTrace();
         fail("Failed to get" + e);
      }
   }

   public void testNodeVisited()
   {
      String rootStr = "/org/jboss/test/data/";

      System.out.println("REGIONS: " + cache_.getRegionManager().dumpRegions());

      for (int i = 0; i < 10; i++)
      {
         String str = rootStr + i;
         Fqn fqn = Fqn.fromString(str);
         try
         {
            cache_.put(fqn, str, str);
         }
         catch (Exception e)
         {
            fail("Failed to insert data" + e);
            e.printStackTrace();
         }
      }

      int period = (wakeupIntervalMillis_ / 2 + 500);
      log("sleeping for " + period + "ms");
      TestingUtil.sleepThread(period);// it really depends the eviction thread time.
      String str = rootStr + "7";
      Fqn fqn = Fqn.fromString(str);
      try
      {
         cache_.get(fqn, str);// just to keep it fresh
         System.out.println("-- sleeping for " + period + "ms");
         TestingUtil.sleepThread(period);// it really depends the eviction thread time.
         cache_.get(fqn, str);// just to keep it fresh
         System.out.println("-- sleeping for " + period + "ms");
         TestingUtil.sleepThread(period);// it really depends the eviction thread time.
         String val = (String) cache_.get(rootStr + "3", rootStr + "3");
         System.out.println("-- val=" + val);
         assertNull("DataNode should be empty ", val);
         val = (String) cache_.get(rootStr + "7", rootStr + "7");
         System.out.println("-- val=" + val);
         assertNotNull("DataNode should not be empty ", val);
         System.out.println("-- sleeping for " + (wakeupIntervalMillis_ + 1000) + "ms");
         TestingUtil.sleepThread(wakeupIntervalMillis_ + 1000);
         val = (String) cache_.get(rootStr + "7", rootStr + "7");
         System.out.println("-- val=" + val);
         assertNull("DataNode should be empty ", val);
      }
      catch (Exception e)
      {
         e.printStackTrace();
         fail("Failed to evict" + e);
      }
   }

   public void testNodeRemoved()
   {
      String rootStr = "/org/jboss/test/data/";
      for (int i = 0; i < 10; i++)
      {
         String str = rootStr + i + "/" + i;
         Fqn fqn = Fqn.fromString(str);
         try
         {
            cache_.put(fqn, str, str);
         }
         catch (Exception e)
         {
            fail("Failed to insert data" + e);
            e.printStackTrace();
         }
      }

      int period = (wakeupIntervalMillis_ / 2 + 500);
      log("period is " + period);
      //      TestingUtil.sleepThread(period);  // it really depends the eviction thread time.
      String str1 = rootStr + "7";
      Fqn fqn1 = Fqn.fromString(str1);
      String str2 = rootStr + "7/7";
      Fqn fqn2 = Fqn.fromString(str2);
      try
      {
         cache_.get(fqn1, str1);// just to keep it fresh
         cache_.get(fqn2, str2);// just to keep it fresh
         TestingUtil.sleepThread(period);// it really depends the eviction thread time.
         cache_.get(fqn1, str1);// just to keep it fresh
         cache_.get(fqn2, str2);// just to keep it fresh
         TestingUtil.sleepThread(period);// it really depends the eviction thread time.
         String val = (String) cache_.get(rootStr + "7/7", rootStr + "7/7");
         assertNotNull("DataNode should not be empty ", val);
         cache_.removeNode(fqn1);
         TestingUtil.sleepThread(wakeupIntervalMillis_ + 500);
         val = (String) cache_.get(rootStr + "7/7", rootStr + "7/7");
         assertNull("DataNode should be empty ", val);
      }
      catch (Exception e)
      {
         e.printStackTrace();
         fail("Failed to evict" + e);
      }
   }

   public void testCompleteRemoval() throws Exception
   {
      String rootStr = "/test/";

      // Add a parent, then a child. LRU will evict the parent,
      // then the child, leaving behind an empty parent
      Fqn<String> parent = Fqn.fromString(rootStr + "parent");
      cache_.put(parent, "key", "value");
      cache_.put(new Fqn<String>(parent, "child"), "key", "value");

      // Give eviction time to run a few times, then confirm parent
      // is completely gone
      TestingUtil.sleepThread((wakeupIntervalMillis_ * 4) + 100);
      assertFalse("Parent completely removed", cache_.getRoot().hasChild(parent));
   }


   class MyPutter extends Thread
   {

      public MyPutter(String name)
      {
         super(name);
      }

      public void run()
      {
         int i = 0;
         final String myName = ROOT_STR + "/test1/node" + getName();
         while (isTrue)
         {
            try
            {
               cache_.put(myName + i++, "value", i);
               sleep(1);
            }
            catch (Throwable e)
            {
               e.printStackTrace();
               if (t1_ex == null)
               {
                  t1_ex = e;
               }
            }
         }
      }
   }


   public void testConcurrentPutAndEvict() throws Exception
   {
      cache_.stop();
      cache_.destroy();
      cache_.getConfiguration().setIsolationLevel(IsolationLevel.REPEATABLE_READ);
      cache_.create();
      cache_.start();
      cache_.put(ROOT_STR + "/concurrentPutAndEvict", "value", 1);

      for (int i = 0; i < 10; i++)
      {
         new MyPutter("Putter" + i).start();
      }

      int counter = 0;
      while (true)
      {
         counter++;
         if (t1_ex != null)
         {
            fail("Exception generated in put() " + t1_ex);
         }
         log("nodes/locks: " + cache_.getNumberOfNodes() + "/" + cache_.getNumberOfLocksHeld());
         TestingUtil.sleepThread(1000);
         if (counter > 10)
         {// run for 10 seconds
            isTrue = false;
            break;
         }
      }
   }

   public void testForEvictionInternalError()
   {
      try
      {
         //       String rootStr = "org/jboss/test/data/";
         String rootStr = "/test/testdata";

         for (int i = 0; i < 10; i++)
         {
            String str = rootStr + i;
            Fqn fqn = Fqn.fromString(str);
            try
            {
               cache_.put(fqn, str, str);
            }
            catch (Exception e)
            {
               fail("Failed to insert data" + e);
               e.printStackTrace();
            }
         }

         // wait for an eviction
         TestingUtil.sleepThread(2 * wakeupIntervalMillis_ + 2000);

         String val = (String) cache_.get(rootStr + "3", rootStr + "3");
         assertNull("DataNode should be empty ", val);

         // reinsert the elements
         for (int i = 0; i < 10; i++)
         {
            String str = rootStr + i;
            Fqn fqn = Fqn.fromString(str);
            try
            {
               cache_.put(fqn, str, str);
            }
            catch (Exception e)
            {
               fail("Failed to insert data" + e);
               e.printStackTrace();
            }
         }

         // clear the root
         cache_.removeNode(Fqn.ROOT);

         // wait for an eviction
         TestingUtil.sleepThread(2 * wakeupIntervalMillis_ + 1000);

         val = (String) cache_.get(rootStr + "3", rootStr + "3");
         assertNull("DataNode should be empty ", val);

      }
      catch (Exception e)
      {
         e.printStackTrace();
         fail("Failed to get" + e);
      }
   }

   void log(String msg)
   {
      System.out.println("-- " + msg);
   }

}
