package org.jboss.cache.profiling;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.config.BuddyReplicationConfig;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.misc.TestingUtil;
import org.jboss.cache.transaction.DummyTransactionManagerLookup;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Test to use with a profiler to profile replication.  To be used in conjunction with ProfileSlaveTest.
 * <p/>
 * Typical usage pattern:
 * <p/>
 * 1.  Start a single test method in ProfileSlaveTest.  This will block until you kill it.
 * 2.  Start the corresponding test in this class, with the same name, in a different JVM, and attached to a profiler.
 * 3.  Profile away!
 * <p/>
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.1.0
 */
@Test(groups = "profiling")
public class ProfileTest extends AbstractProfileTest
{
   /*
      Test configuration options
    */
   protected static final long DURATION = 2 * 60 * 1000;
   protected static final int NUM_THREADS = 10;
   protected static final int MAX_RANDOM_SLEEP_MILLIS = 100;
   protected static final int MAX_DEPTH = 8;
   protected static final int MAX_OVERALL_NODES = 100;
   protected static final int WARMUP_LOOPS = 10000;

   private List<Fqn> fqns = new ArrayList<Fqn>(MAX_OVERALL_NODES);
   private Random r = new Random();

   private Log log = LogFactory.getLog(ProfileTest.class);

   public void testReplSync() throws Exception
   {
      runCompleteTest();
   }

   public void testReplAsync() throws Exception
   {
      cache.getConfiguration().setCacheMode(Configuration.CacheMode.REPL_ASYNC);
      runCompleteTest();
   }

   public void testReplSyncOptimistic() throws Exception
   {
      cache.getConfiguration().setNodeLockingScheme(Configuration.NodeLockingScheme.OPTIMISTIC);
      cache.getConfiguration().setTransactionManagerLookupClass(DummyTransactionManagerLookup.class.getName());
      runCompleteTest();
   }

   public void testReplAsyncOptimistic() throws Exception
   {
      cache.getConfiguration().setCacheMode(Configuration.CacheMode.REPL_ASYNC);
      cache.getConfiguration().setNodeLockingScheme(Configuration.NodeLockingScheme.OPTIMISTIC);
      cache.getConfiguration().setTransactionManagerLookupClass(DummyTransactionManagerLookup.class.getName());
      runCompleteTest();
   }

   public void testReplSyncBR() throws Exception
   {
      BuddyReplicationConfig brc = new BuddyReplicationConfig();
      brc.setEnabled(true);
      cache.getConfiguration().setBuddyReplicationConfig(brc);
      testReplSync();
   }

   public void testReplAsyncBR() throws Exception
   {
      BuddyReplicationConfig brc = new BuddyReplicationConfig();
      brc.setEnabled(true);
      cache.getConfiguration().setBuddyReplicationConfig(brc);
      testReplAsync();
   }

   public void testReplSyncOptBR() throws Exception
   {
      BuddyReplicationConfig brc = new BuddyReplicationConfig();
      brc.setEnabled(true);
      cache.getConfiguration().setBuddyReplicationConfig(brc);
      testReplSyncOptimistic();
   }

   public void testReplAsyncOptBR() throws Exception
   {
      BuddyReplicationConfig brc = new BuddyReplicationConfig();
      brc.setEnabled(true);
      cache.getConfiguration().setBuddyReplicationConfig(brc);
      testReplAsyncOptimistic();
   }

   private void runCompleteTest() throws Exception
   {
      init();
      startup();
      warmup();
      doTest();
   }

   /**
    * Thr following test phases can be profiled individually using triggers in JProfiler.
    */

   protected void init()
   {
      long startTime = System.currentTimeMillis();
      log.warn("Starting init() phase");
      fqns.clear();
      for (int i = 0; i < MAX_OVERALL_NODES; i++)
      {
         Fqn fqn = createRandomFqn(r);
         while (fqns.contains(fqn)) fqn = createRandomFqn(r);
         if (i % 10 == 0)
         {
            log.warn("Generated " + i + " fqns");
         }
         fqns.add(fqn);
      }
      System.gc();
      long duration = System.currentTimeMillis() - startTime;
      log.warn("Finished init() phase.  " + printDuration(duration));
   }

   private Fqn createRandomFqn(Random r)
   {
      String s = "/";
      int depth = r.nextInt(MAX_DEPTH);
      for (int i = 0; i < depth; i++)
      {
         s += r.nextInt(Integer.MAX_VALUE) + "/";
      }

      return Fqn.fromString(s);
   }


   private void startup()
   {
      long startTime = System.currentTimeMillis();
      log.warn("Starting cache");
      cache.start();
      long duration = System.currentTimeMillis() - startTime;
      log.warn("Started cache.  " + printDuration(duration));
   }

   private void warmup()
   {
      long startTime = System.currentTimeMillis();
      log.warn("Starting warmup");
      // creates all the Fqns since this can be expensive and we don't really want to measure this (for now)
      for (Fqn f : fqns)
      {
         // this will create the necessary nodes.
         cache.put(f, Collections.emptyMap());
      }

      // loop through WARMUP_LOOPS gets and puts for JVM optimisation
      for (int i = 0; i < WARMUP_LOOPS; i++)
      {
         Fqn f = fqns.get(r.nextInt(MAX_OVERALL_NODES));
         cache.get(f, "");
         cache.put(f, "k", "v");
         cache.remove(f, "k");
      }

      long duration = System.currentTimeMillis() - startTime;
      log.warn("Finished warmup.  " + printDuration(duration));
   }

   private void doTest() throws Exception
   {
      ExecutorService exec = Executors.newFixedThreadPool(NUM_THREADS);
      long end = System.currentTimeMillis() + DURATION;
      long startTime = System.currentTimeMillis();
      log.warn("Starting test");
      while (System.currentTimeMillis() < end)
      {
         exec.execute(new Runnable()
         {
            public void run()
            {
               String k = getRandomString();
               String v = getRandomString();
               cache.put(fqns.get(r.nextInt(MAX_OVERALL_NODES)), k, v);
            }
         });
         TestingUtil.sleepRandom(MAX_RANDOM_SLEEP_MILLIS);
      }
      log.warn("Finished generating runnables; awaiting executor completion");
      // wait for executors to complete!
      exec.shutdown();
      exec.awaitTermination(10000, TimeUnit.MILLISECONDS);
      long duration = System.currentTimeMillis() - startTime;
      log.warn("Finished test.  " + printDuration(duration));
   }

   private String getRandomString()
   {
      StringBuilder sb = new StringBuilder();
      int len = r.nextInt(10);

      for (int i = 0; i < len; i++)
      {
         sb.append((char) (63 + r.nextInt(26)));
      }
      return sb.toString();
   }

   private String printDuration(long duration)
   {
      if (duration > 2000)
      {
         double dSecs = ((double) duration / (double) 1000);
         return "Duration: " + dSecs + " seconds";
      }
      else
      {
         return "Duration: " + duration + " millis";
      }
   }
}
